package AITB_2022.tutorial.module.complex;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.info.*;
import adf.core.component.module.complex.HumanDetector;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import adf.core.component.module.algorithm.Clustering;
import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;

import static adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam.ACTION_LOAD;
import static java.lang.Double.POSITIVE_INFINITY;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;


public class AITHumanDetector extends HumanDetector
{
    private EntityID result;
    private Clustering clustering;
    private MessageManager messageManager;
    private Logger logger;
    private EntityID lastID=null;

    public AITHumanDetector(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);

        this.clustering = mm.getModule("SampleHumanDetector.Clustering", "adf.impl.module.algorithm.KMeansClustering");
        this.registerModule(this.clustering);

        this.logger = DefaultLogger.getLogger(agentInfo.me());
    }

    //行動対象をreturnするメソッド
    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if (this.getCountUpdateInfo() > 1) {
            return this;
        }
        this.messageManager = messageManager;

        return this;
    }

    //行動対象を計算するメソッド
    @Override
    public HumanDetector calc()
    {
        //このエージェントが担当するクラスタの番号を取得し，idxに代入
        int idx = this.clustering.getClusterIndex(this.agentInfo.getID());
        //idx番のクラスタに属するEntityをすべて取得し，myClusterEntitiesに代入
        Collection<StandardEntity> myClusterEntities = this.clustering.getClusterEntities(idx);
        //シミュレーションの1ステップ目のみで，このエージェントが担当するクラスタに属するEntityを表示
        if (this.agentInfo.getTime() == 1) logger.debug(myClusterEntities);

        EntityID me = this.agentInfo.getID();
        Collection<EntityID> ids = this.worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN);

        EntityID nearestID = null;
        double nearestDistance = Double.POSITIVE_INFINITY;

        for (EntityID id : ids)
        {
            StandardEntity entity = this.worldInfo.getEntity(id);
            Civilian civilian = (Civilian) entity;
            boolean isAlive = civilian.isHPDefined() && civilian.getHP() > 0;
            if (!isAlive) continue;
            boolean isBury = civilian.isBuriednessDefined() && civilian.getBuriedness() > 0;
            if (isBury)
            {
                double distance = this.worldInfo.getDistance(me, id);
                if (distance < nearestDistance)
                {
                    nearestID = id;
                    nearestDistance = distance;
                }
            }
        }
        this.result = nearestID;

        if(lastID != null && lastID != this.result){
            Civilian civ = (Civilian) this.worldInfo.getEntity(this.result);
            if (civ != null) {
                messageManager.addMessage(new MessageCivilian(false,civ),false);
            }
        }
        lastID = this.result;
        return this;
    }
}
