package AITB_2022.tutorial.module.complex;
import adf.core.agent.info.*;
import adf.core.component.module.complex.Search;
import adf.core.component.module.algorithm.Clustering;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;

public class AITPFSearch extends Search {
    private EntityID result;
    private Random random = new Random(1);
    private Clustering clustering;
    private List<EntityID> list;
    private Set<EntityID> searchedSet = new HashSet<>();
    private List<EntityID> rmlist = new ArrayList<>();

    public AITPFSearch(
        AgentInfo ai, WorldInfo wi, ScenarioInfo si,
        ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);

        String clusteringKey = "SampleSearch.Clustering.Police";
        this.clustering =
            mm.getModule(clusteringKey,
                    "adf.sample.module.algorithm.SampleKMeans");
        this.registerModule(this.clustering);
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    @Override
    public Search calc()
    {
        EntityID me = this.agentInfo.getID();
        int idx = this.clustering.getClusterIndex(me);
        if (idx == -1) {
        return this;
        }

        Collection<EntityID> cluster =
                this.clustering.getClusterEntityIDs(idx);
        if (cluster == null) {
        return this;
        }

        // エージェントが現在位置する建物や道路のIDを取得する
        EntityID position = this.agentInfo.getPosition();

        if(this.result == null) {
            list = new ArrayList<>(cluster);
            // 建物以外を除く
            for (EntityID rm : list) {
                StandardEntity entity = this.worldInfo.getEntity(rm);
                if(entity instanceof Building) {
                    this.rmlist.add(rm);
                }
            }

            for (EntityID l : rmlist) {
                list.remove(l);
            }
            // 擬似乱数を用いてランダムに選択します
            list.remove(this.result);
            int n = list.size();
            int r = random.nextInt(n);
            this.result = list.get(r);
        }

        // 対象に辿り着くまで，対象を変更しない
        if (!this.getTarget().equals(position)) {
        return this;
        }

        // 探索済みの (辿り着いたことがある) 対象を候補から除外する
        if (this.getTarget().equals(position)) {
        this.searchedSet.add(this.getTarget());
        list.removeAll(this.searchedSet);

        // すべての対象が探索済みの場合，候補をリセットする
        if (list.size() <= 0) {
            this.searchedSet.clear();
            list = new ArrayList<>(cluster);
            for (EntityID l : rmlist) {
                list.remove(l);
            }
        }

        // 擬似乱数を用いてランダムに選択します
        int n = list.size();
        int r = this.random.nextInt(n);
        this.result = list.get(r);

        return this;
        }

        return this;
    }
}