package AITB_2022.tutorial.module.fb_suzuki;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.centralized.CommandAmbulance;
import adf.core.agent.communication.standard.bundle.information.MessageCivilian;
import adf.core.agent.communication.standard.bundle.information.MessageRoad;
import adf.core.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.core.agent.info.*;
import adf.core.component.module.complex.HumanDetector;
import adf.core.component.communication.CommunicationMessage;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.develop.DevelopData;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import java.util.*;
import adf.core.component.module.algorithm.Clustering;
import adf.core.debug.DefaultLogger;
import org.apache.log4j.Logger;

import static adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam.ACTION_LOAD;
import static java.lang.Double.POSITIVE_INFINITY;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;

//timeがステップ数 targetが市民のエンティティID 
// 行動(掘り起こし)対象(市民)を決定するクラス
public class AITFBHumanDetector extends HumanDetector
{
    private EntityID result;
    private Clustering clustering;
    private MessageManager messageManager;
    private Logger logger;
    private EntityID lastID = null;
    private double lastX = -1.0;//1step前のエージェント自身のX座標
    private double lastY = -1.0;//1step前のエージェント自身のY座標
    private int police_coll_flag = 0;//土木隊を呼んだときのクールタイム

    //通信用のアクション番号
    final private int HELP_BLOCKADE = 6;//瓦礫で動けない状態

    public AITFBHumanDetector(
            AgentInfo ai, WorldInfo wi, ScenarioInfo si,
            ModuleManager mm, DevelopData dd)
    {
        super(ai, wi, si, mm, dd);
        this.clustering = mm.getModule("SampleHumanDetector.Clustering", "adf.impl.module.algorithm.KMeansClustering");
        this.registerModule(this.clustering);

        this.logger = DefaultLogger.getLogger(agentInfo.me());
    }

    //行動対象をreturnするメソッド
    @Override
    public EntityID getTarget()
    {
        return this.result;
    }
    
    @Override
    public HumanDetector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        if (this.getCountUpdateInfo() > 1) {
            return this;
        }
        this.messageManager = messageManager;
        // そのステップで受信した(前ステップで送信された)メッセージ一覧
        List<CommunicationMessage> messages = messageManager.getReceivedMessageList();
        this.logger.debug("Time:"+this.agentInfo.getTime()+" mess: "+messages);

        return this;
    }

    //行動対象を計算するメソッド
    @Override
    public HumanDetector calc()
    {
        //このエージェントが担当するクラスタの番号を取得し，idxに代入
        int idx = this.clustering.getClusterIndex(this.agentInfo.getID());
        //idx番のクラスタに属するEntityをすべて取得し，myClusterEntitiesに代入
        Collection<StandardEntity> myClusterEntities = this.clustering.getClusterEntities(idx);
        //シミュレーションの1ステップ目のみ、このエージェントが担当するクラスタに属するEntityを表示
        if (this.agentInfo.getTime() == 1) logger.debug(myClusterEntities);
        
        EntityID me = this.agentInfo.getID();//自分のエージェントのIDを取得
        Collection<EntityID> ids = this.worldInfo.getEntityIDsOfType(StandardEntityURN.CIVILIAN);//知覚している市民のEntityIDコレクションを取得

        //Collection<StandardEntity> entitys = this.worldInfo.getEntitiesOfType(StandardEntityURN.CIVILIAN);//知覚している市民のEntityIDコレクションを取得
        //this.logger.debug("time"+this.agentInfo.getTime()+" Civilian_entities "+entitys);
        
        //最初の3stepは消防隊が止まってる？から3stepまでは行わない
        if(this.agentInfo.getTime() > 3){
            this.message_police((FireBrigade)this.agentInfo.me(),this.agentInfo.getPosition());//自分が瓦礫で止まっているかを判定して真なら土木隊に送信
        }

        EntityID nearestID = null;//一番近い市民のIDを格納する変数 初期値はnull
        double nearestDistance = Double.POSITIVE_INFINITY;//nearesDistanceに初期値としてDoubleで最も大きい値を代入

        for (EntityID id : ids)//拡張forで市民のEntityIDをまわす
        {
            StandardEntity entity = this.worldInfo.getEntity(id);//idからentityを取得
            Civilian civilian = (Civilian) entity;//ダウンキャスト
            boolean isAlive = civilian.isHPDefined() && civilian.getHP() > 0;//生きてたらtrue
            
            //クラスタ内のEntityに市民の場所のEntityが含まれていなかったらスルー
            if(!myClusterEntities.contains(this.worldInfo.getEntity(civilian.getPosition()))) continue;
            //logger.debug("time"+this.agentInfo.getTime()+" EntityID "+civilian.getPosition());

            //死んでたらスルー
            if (!isAlive) continue;
            boolean isBury = civilian.isBuriednessDefined() && civilian.getBuriedness() > 0;//埋まってたらtrue
            //埋まってら処理する
            if (isBury)
            {
                double distance = this.worldInfo.getDistance(me, id);
                if (distance < nearestDistance)
                {
                    nearestID = id;
                    nearestDistance = distance;
                }
            }
        }
        this.result = nearestID;

        if(lastID != null && lastID != this.result){
            Civilian civ = (Civilian) this.worldInfo.getEntity(this.result);
            if (civ != null) {
                messageManager.addMessage(new MessageCivilian(false,civ),false);
            }
        }//getExcutedAction
        lastID = this.result;

        return this;
    }
    
    // 現在、瓦礫で身動きが取れないかを判断し、真なら土木隊に送信する処理
    public void message_police(FireBrigade fbMeEntity, EntityID nowPositionID){//第一引数は自身のEntity(FireBrigade), 第二引数は現在場所のID
        //エージェント自身が現在居る場所のエンティティをIDから取得
        StandardEntity nowPositionEntity = this.worldInfo.getEntity(nowPositionID);

        double nowX = this.agentInfo.getX();//現在のX座標を取得
        double nowY = this.agentInfo.getY();//現在のY座標を取得
        this.logger.debug("Time:"+this.agentInfo.getTime()+" nowPositionEntity "+nowPositionEntity);

        double compDistanceX = Math.abs((nowX - this.lastX));//現在のX座標と1step前のX座標との差
        double compDistanceY = Math.abs((nowY - this.lastY));//現在のY座標と1step前のY座標との差

        this.logger.debug("nowX "+nowX+" lastX "+lastX+" compDistanceX "+compDistanceX);
        this.logger.debug("nowY "+nowY+" lastY "+lastY+" compDistanceY "+compDistanceY);

        //もしエージェントが現在いる場所が道路で、1step前と場所が全く変わっていなくて、土木隊を呼ぶクールタイムが0のとき
        if((nowPositionEntity instanceof Road) && ((compDistanceX < 3.0)&&(compDistanceY < 3.0)) && (this.police_coll_flag == 0)){
            this.police_coll_flag = 10;//土木隊を呼ぶクールタイムを設定
            this.logger.debug("現在、Road居てさっきとエージェントの場所が変わってなくてクールタイム中でもない。つまり瓦礫で止まってます");

            //消防隊が埋まったという情報を送信
            messageManager.addMessage(new MessageFireBrigade(false,fbMeEntity,HELP_BLOCKADE,nowPositionID),false);

        }else if(this.police_coll_flag != 0){//もし土木隊を呼ぶクールタイムが0でないなら
            this.police_coll_flag -= 1;//土木隊を呼ぶクールタイムを減少させる
        }
        this.lastX = nowX;//今のXをlastXに代入
        this.lastY = nowY;//今のYをlastYに代入
    }
}