package AITB_2022.tutorial.module.fb_suzuki;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_CENTRE;
import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.BUILDING;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_OFFICE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.algorithm.PathPlanning;
import adf.core.component.module.complex.Search;
import adf.core.debug.DefaultLogger;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.log4j.Logger;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

public class AITFBSearch extends Search {

  private PathPlanning pathPlanning;
  private Clustering clustering;//クラスタ

  private EntityID result;//探索対象
  private Collection<EntityID> unsearchedBuildingIDs;//まだ探索していないエンティティのID
  private Logger logger;

  public AITFBSearch(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    this.logger = DefaultLogger.getLogger(agentInfo.me());
    this.unsearchedBuildingIDs = new HashSet<>();

    StandardEntityURN agentURN = ai.me().getStandardURN();
    if (agentURN == AMBULANCE_TEAM) {
      this.pathPlanning = moduleManager.getModule(
          "SampleSearch.PathPlanning.Ambulance",
          "adf.impl.module.algorithm.DijkstraPathPlanning");
      this.clustering = moduleManager.getModule(
          "SampleSearch.Clustering.Ambulance",
          "adf.impl.module.algorithm.KMeansClustering");
    } else if (agentURN == FIRE_BRIGADE) {
      this.pathPlanning = moduleManager.getModule(
          "SampleSearch.PathPlanning.Fire",
          "adf.impl.module.algorithm.DijkstraPathPlanning");
      this.clustering = moduleManager.getModule("SampleSearch.Clustering.Fire",
          "adf.impl.module.algorithm.KMeansClustering");
    } else if (agentURN == POLICE_FORCE) {
      this.pathPlanning = moduleManager.getModule(
          "SampleSearch.PathPlanning.Police",
          "adf.impl.module.algorithm.DijkstraPathPlanning");
      this.clustering = moduleManager.getModule(
          "SampleSearch.Clustering.Police",
          "adf.impl.module.algorithm.KMeansClustering");
    }
    registerModule(this.clustering);
    registerModule(this.pathPlanning);
  }


  @Override
  public Search updateInfo(MessageManager messageManager) {//unsearchedBuildingIDsの更新処理
    //logger.debug("Time:" + agentInfo.getTime() + " updateInfo");
    super.updateInfo(messageManager);//継承親のupdateInfoを実行

    this.unsearchedBuildingIDs
        .removeAll(this.worldInfo.getChanged().getChangedEntities());//自身が知覚しているエンティティのIDを未探索エンティティIDコレクションから消す
    if (this.unsearchedBuildingIDs.isEmpty()) {//もし全て探索し終わったなら
      this.reset();//unsearchedBuildingIDsをリセット
      this.unsearchedBuildingIDs
          .removeAll(this.worldInfo.getChanged().getChangedEntities());//もう一度、自身が知覚しているエンティティのIDを消しておく
    }
    //logger.debug("Time:" + agentInfo.getTime()+" updateInfo"+" unsearchedBuildingIDs: " + this.unsearchedBuildingIDs);
    String a = "";
    for(EntityID id : this.unsearchedBuildingIDs){
      a += this.worldInfo.getEntity(id)+" ";
    }
    logger.debug("Time:" + agentInfo.getTime()+" updateInfo"+" unsearchedBuildings: " + a);
    return this;
  }


  @Override
  public Search calc() {//探索対象を見つける処理
    this.result = null;
    if (unsearchedBuildingIDs.isEmpty())//もしunsearchedBuildingIDsに要素がないなら
      return this;

    logger.debug("Time:"+this.agentInfo.getTime()+" calc " + "unsearchedBuildingIDs: " + unsearchedBuildingIDs);
    this.pathPlanning.setFrom(this.agentInfo.getPosition());
    this.pathPlanning.setDestination(this.unsearchedBuildingIDs);
    List<EntityID> path = this.pathPlanning.calc().getResult();
    String a = "";
    for(EntityID id : path){
      a += this.worldInfo.getEntity(id)+" ";
    }
    logger.debug("best path is: " + path);
    logger.debug("calc "+"pathentity: " + a);
    if (path != null && path.size() > 2) {
      this.result = path.get(path.size() - 3);
    } else if (path != null && path.size() > 0) {
      this.result = path.get(path.size() - 1);
    }
    logger.debug("chose: " + result);
    return this;
  }


  private void reset() {
    this.logger.debug("time:"+this.agentInfo.getTime()+"reset " + "unsearchBuildingIDsをリセットしました");
    this.unsearchedBuildingIDs.clear();//全ての要素をコレクションから消す
    //このエージェントが担当するクラスタの番号を取得し，clusterIndexに代入
    int clusterIndex = this.clustering.getClusterIndex(this.agentInfo.getID());
    //clusterIndex番のクラスタに属するEntityをすべて取得し，clusterEntitiesに代入
    Collection<StandardEntity> clusterEntities = this.clustering
        .getClusterEntities(clusterIndex);
    if (clusterEntities != null && clusterEntities.size() > 0) {//clusterEntitiesがnullでなく要素があるとき
      for (StandardEntity entity : clusterEntities) {//clusterEntitiesの拡張for
        if (entity instanceof Building && entity.getStandardURN() != REFUGE) {//entityがBuildingクラスのインスタンスかつentityが持つイーナムがREFUGEでないとき
          this.unsearchedBuildingIDs.add(entity.getID());//entityのIDをunsearchBuildingIDsに追加
        }
      }
    } else {//clusterEntitiesがnullもしくは要素がないとき
      this.unsearchedBuildingIDs
          .addAll(this.worldInfo.getEntityIDsOfType(BUILDING, GAS_STATION,
              AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE));//エージェントが把握しているBUILDING,GAS_STATION,AMBULANCE_CENTRE,FIRE_STATION,POLICE_OFFICEのEntityIDコレクションの要素を全てunsearchedBuildingIDsに追加
    }
  }


  @Override
  public EntityID getTarget() {
    return this.result;
  }
}